from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import LargeBinary
from sqlalchemy import orm
from sqlalchemy.orm import relationship
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_plugin_chat._private.PluginNames import chatTuplePrefix
from .ChatUserTuple import ChatUserTuple
from .DeclarativeBase import DeclarativeBase
from .MessageTuple import MessageTuple


@addTupleType
class MessageReadPayloadTuple(DeclarativeBase, Tuple):
    __tupleType__ = chatTuplePrefix + "MessageReadPayloadTuple"
    __tablename__ = "MessageReadPayloadTuple"

    id = Column(Integer, primary_key=True, autoincrement=True)

    #: Foreign key to a Message
    messageId = Column(
        Integer, ForeignKey(MessageTuple.id, ondelete="CASCADE"), nullable=False
    )
    message = relationship(MessageTuple)

    #: Foreign key to a ChatUser
    chatUserId = Column(
        Integer, ForeignKey(ChatUserTuple.id, ondelete="CASCADE"), nullable=True
    )
    chatUser = relationship(ChatUserTuple)

    onReadPayload = Column(LargeBinary, nullable=False)

    __table_args__ = (
        Index("idx_ChatPayloads", messageId, chatUserId, unique=False),
    )

    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)



"""Initial Schema

Peek Plugin Database Migration Script

Revision ID: 5288d02e94e4
Revises: 
Create Date: 2017-04-09 12:10:02.829698

"""

# revision identifiers, used by Alembic.
revision = "5288d02e94e4"
down_revision = None
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "ChatTuple",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("lastActivity", sa.DateTime(), nullable=False),
        sa.Column("usersKey", sa.String(length=4000), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        schema="pl_chat",
    )
    op.create_index(
        "idx_Chat_usersKey",
        "ChatTuple",
        ["usersKey"],
        unique=True,
        schema="pl_chat",
    )
    op.create_table(
        "Setting",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        schema="pl_chat",
    )
    op.create_table(
        "ChatUserTuple",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("chatId", sa.Integer(), nullable=False),
        sa.Column("userId", sa.String(length=2000), nullable=False),
        sa.Column("isUserExternal", sa.Boolean(), nullable=False),
        sa.Column("lastReadDate", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["chatId"], ["pl_chat.ChatTuple.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="pl_chat",
    )
    op.create_index(
        "idx_ChatUser_chatId",
        "ChatUserTuple",
        ["chatId"],
        unique=False,
        schema="pl_chat",
    )
    op.create_index(
        "idx_ChatUser_userId",
        "ChatUserTuple",
        ["userId"],
        unique=False,
        schema="pl_chat",
    )
    op.create_table(
        "MessageTuple",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("chatId", sa.Integer(), nullable=False),
        sa.Column("message", sa.String(length=2000), nullable=False),
        sa.Column("priority", sa.Integer(), nullable=False),
        sa.Column("fromUserId", sa.String(length=40), nullable=False),
        sa.Column("dateTime", sa.DateTime(), nullable=False),
        sa.Column("onReadPayload", sa.LargeBinary(), nullable=True),
        sa.ForeignKeyConstraint(
            ["chatId"], ["pl_chat.ChatTuple.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="pl_chat",
    )
    op.create_index(
        "idx_ChatMsgTuple_chatId",
        "MessageTuple",
        ["chatId"],
        unique=False,
        schema="pl_chat",
    )
    op.create_index(
        "idx_ChatMsgTuple_dateTime",
        "MessageTuple",
        ["dateTime"],
        unique=False,
        schema="pl_chat",
    )
    op.create_table(
        "SettingProperty",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("settingId", sa.Integer(), nullable=False),
        sa.Column("key", sa.String(length=50), nullable=False),
        sa.Column("type", sa.String(length=16), nullable=True),
        sa.Column("int_value", sa.Integer(), nullable=True),
        sa.Column("char_value", sa.String(), nullable=True),
        sa.Column("boolean_value", sa.Boolean(), nullable=True),
        sa.ForeignKeyConstraint(
            ["settingId"],
            ["pl_chat.Setting.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="pl_chat",
    )
    op.create_index(
        "idx_SettingProperty_settingId",
        "SettingProperty",
        ["settingId"],
        unique=False,
        schema="pl_chat",
    )
    op.create_table(
        "MessageReadPayloadTuple",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("messageId", sa.Integer(), nullable=False),
        sa.Column("chatUserId", sa.Integer(), nullable=True),
        sa.Column("onReadPayload", sa.LargeBinary(), nullable=False),
        sa.ForeignKeyConstraint(
            ["chatUserId"], ["pl_chat.ChatUserTuple.id"], ondelete="NO ACTION"
        ),
        sa.ForeignKeyConstraint(
            ["messageId"], ["pl_chat.MessageTuple.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
        schema="pl_chat",
    )
    op.create_index(
        "idx_ChatPayloads",
        "MessageReadPayloadTuple",
        ["messageId", "chatUserId"],
        unique=False,
        schema="pl_chat",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        "idx_ChatPayloads",
        table_name="MessageReadPayloadTuple",
        schema="pl_chat",
    )
    op.drop_table("MessageReadPayloadTuple", schema="pl_chat")
    op.drop_index(
        "idx_SettingProperty_settingId",
        table_name="SettingProperty",
        schema="pl_chat",
    )
    op.drop_table("SettingProperty", schema="pl_chat")
    op.drop_index(
        "idx_ChatMsgTuple_dateTime", table_name="MessageTuple", schema="pl_chat"
    )
    op.drop_index(
        "idx_ChatMsgTuple_chatId", table_name="MessageTuple", schema="pl_chat"
    )
    op.drop_table("MessageTuple", schema="pl_chat")
    op.drop_index(
        "idx_ChatUser_userId", table_name="ChatUserTuple", schema="pl_chat"
    )
    op.drop_index(
        "idx_ChatUser_chatId", table_name="ChatUserTuple", schema="pl_chat"
    )
    op.drop_table("ChatUserTuple", schema="pl_chat")
    op.drop_table("Setting", schema="pl_chat")
    op.drop_index("idx_Chat_usersKey", table_name="ChatTuple", schema="pl_chat")
    op.drop_table("ChatTuple", schema="pl_chat")
    # ### end Alembic commands ###



from peek_plugin_chat._private.PluginNames import chatTuplePrefix
from vortex.Tuple import addTupleType, TupleField
from vortex.TupleAction import TupleActionABC


@addTupleType
class SendMessageActionTuple(TupleActionABC):
    __tupleType__ = chatTuplePrefix + "SendMessageActionTuple"

    chatId = TupleField(typingType=int)
    fromUserId = TupleField(typingType=str)
    message = TupleField(typingType=str)
    priority = TupleField(typingType=int)



import logging
from typing import List

from reactivex.subject import Subject
from twisted.internet import reactor

from peek_plugin_chat._private.server.controller.MainController import (
    MainController,
)
from peek_plugin_chat._private.storage.ChatTuple import ChatTuple
from peek_plugin_chat._private.storage.MessageTuple import MessageTuple
from peek_plugin_chat.server.ChatApiABC import (
    ChatApiABC,
    NewMessage,
    ReceivedMessage,
)
from vortex.DeferUtil import yesMainThread

logger = logging.getLogger(__name__)


class ChatApi(ChatApiABC):
    def __init__(self, ormSessionCreator):
        self._ormSessionCreator = ormSessionCreator
        self._mainController = None

        self._observablesByUserId = {}

    def setMainController(self, mainController: MainController):
        self._mainController = mainController

    def shutdown(self):
        self._mainController = None
        for subject in self._observablesByUserId.values():
            subject.dispose()

    def notifyOfReceivedMessage(self, chat: ChatTuple, message: MessageTuple):
        receivedMessage = ReceivedMessage(
            fromUserId=message.fromUserId,
            allUserIds=[cu.userId for cu in chat.users],
            message=message.message,
            priority=message.priority,
        )

        for user in chat.users:
            # checking if it's external only isn't really necessary
            # other APIs could observe any users message.
            if not user.isUserExternal:
                continue

            # We don't notify the sender
            if user.userId == message.fromUserId:
                continue

            # And if no ones watching it, there is nothing to do.
            if user.userId not in self._observablesByUserId:
                continue

            self._observablesByUserId[user.userId].on_next(receivedMessage)

    def sendMessage(self, newMessage: NewMessage) -> None:
        yesMainThread()
        return self._mainController.sendMessageFromExternalUser(newMessage)

    def createChat(self, fromExtUserId: str, toUserIds: List[str]) -> None:
        yesMainThread()
        return self._mainController.createChat(fromExtUserId, toUserIds)

    def receiveMessages(self, toExtUserId: str) -> Subject:
        yesMainThread()
        if toExtUserId not in self._observablesByUserId:
            self._observablesByUserId[toExtUserId] = Subject()
        return self._observablesByUserId[toExtUserId]



from sqlalchemy import Column
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import orm
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import ForeignKey
from sqlalchemy.sql.sqltypes import Boolean
from sqlalchemy.sql.sqltypes import DateTime
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_plugin_chat._private.PluginNames import chatTuplePrefix
from peek_plugin_chat._private.storage.ChatTuple import ChatTuple
from peek_plugin_chat._private.storage.DeclarativeBase import DeclarativeBase


@addTupleType
class ChatUserTuple(DeclarativeBase, Tuple):
    __tupleType__ = chatTuplePrefix + "ChatUserTuple"
    __tablename__ = "ChatUserTuple"

    id = Column(Integer, primary_key=True, autoincrement=True)

    #: Foreign key to a chat
    chatId = Column(
        Integer, ForeignKey(ChatTuple.id, ondelete="CASCADE"), nullable=False
    )
    chat = relationship(ChatTuple, overlaps="users")

    #: The userId of a user in the chat
    userId = Column(String(2000), nullable=False)
    isUserExternal = Column(Boolean, nullable=False)

    #: Last Read Date
    lastReadDate = Column(DateTime(True), nullable=False)
    # If ChatUserTuple.lastReadDate < ChatTuple.lastActivity then we have unread
    # messages.

    # #:  User Name, to be populated before sending to the UI
    # userName = TupleField(defaultValue="Unknown")

    __table_args__ = (
        Index("idx_ChatUser_userId", userId, unique=False),
        Index("idx_ChatUser_chatId", chatId, unique=False),
    )

    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)



from vortex.handler.TupleActionProcessor import TupleActionProcessor

from peek_plugin_chat._private.PluginNames import chatFilt
from peek_plugin_chat._private.PluginNames import chatActionProcessorName
from .controller.MainController import MainController


def makeTupleActionProcessorHandler(mainController: MainController):
    processor = TupleActionProcessor(
        tupleActionProcessorName=chatActionProcessorName,
        additionalFilt=chatFilt,
        defaultDelegate=mainController,
    )
    return processor



from peek_plugin_base.PeekVortexUtil import peekServerName
from peek_plugin_chat._private.PluginNames import chatFilt
from peek_plugin_chat._private.PluginNames import chatObservableName
from vortex.handler.TupleDataObservableProxyHandler import (
    TupleDataObservableProxyHandler,
)


def makeDeviceTupleDataObservableProxy():
    return TupleDataObservableProxyHandler(
        observableName=chatObservableName,
        proxyToVortexName=peekServerName,
        additionalFilt=chatFilt,
    )



def loadPrivateTuples():
    """Load Private Tuples

    In this method, we load the private tuples.
    This registers them so the Vortex can reconstructed them from
    serialised data.

    """
    from . import SendMessageActionTuple

    SendMessageActionTuple.__unused = False

    from . import CreateChatActionTuple

    CreateChatActionTuple.__unused = False

    from . import ChatUserReadActionTuple

    ChatUserReadActionTuple.__unused = False



from datetime import datetime

from peek_plugin_chat._private.PluginNames import chatTuplePrefix
from vortex.Tuple import addTupleType, TupleField
from vortex.TupleAction import TupleActionABC


@addTupleType
class ChatUserReadActionTuple(TupleActionABC):
    __tupleType__ = chatTuplePrefix + "ChatUserReadActionTuple"

    chatUserId = TupleField(typingType=int)
    readDateTime = TupleField(typingType=datetime)



from sqlalchemy import Column
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import orm
from sqlalchemy.orm import relationship
from sqlalchemy.sql.sqltypes import DateTime
from sqlalchemy.sql.sqltypes import String
from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from peek_plugin_chat._private.PluginNames import chatTuplePrefix
from peek_plugin_chat._private.storage.DeclarativeBase import DeclarativeBase


@addTupleType
class ChatTuple(DeclarativeBase, Tuple):
    __tupleType__ = chatTuplePrefix + "ChatTuple"
    __tablename__ = "ChatTuple"

    # This will include the users relationship when serialising the data
    __fieldNames__ = ["users"]

    id = Column(Integer, primary_key=True, autoincrement=True)

    # Message details
    lastActivity = Column(DateTime(True), nullable=False)

    #: Unique users key, A comma separated string of the users in this chat
    #  This is used to ensure there is only one chat for these user combinations.
    usersKey = Column(String(4000), nullable=False)

    # Use a TupleField instead of a relationship so we can decide when it will
    # include the messages or not.
    messages = TupleField()  # relationship("Message.chatId")

    users = relationship("ChatUserTuple", lazy="joined")

    __table_args__ = (Index("idx_Chat_usersKey", usersKey, unique=True),)

    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)



import logging
from typing import Optional

from peek_plugin_base.server.PluginLogicEntryHookABC import (
    PluginLogicEntryHookABC,
)
from peek_plugin_base.server.PluginServerStorageEntryHookABC import (
    PluginServerStorageEntryHookABC,
)
from peek_plugin_chat._private.server.ChatApi import ChatApi
from peek_plugin_chat._private.server.TupleActionProcessor import (
    makeTupleActionProcessorHandler,
)
from peek_plugin_chat._private.server.TupleDataObservable import (
    makeTupleDataObservableHandler,
)
from peek_plugin_chat._private.server.admin_backend import (
    makeAdminBackendHandlers,
)
from peek_plugin_chat._private.server.controller.MainController import (
    MainController,
)
from peek_plugin_chat._private.server.controller.TaskController import (
    TaskController,
)
from peek_plugin_chat._private.storage import DeclarativeBase, loadStorageTuples
from peek_plugin_chat._private.tuples import loadPrivateTuples
from peek_plugin_chat.tuples import loadPublicTuples
from peek_plugin_inbox.server.InboxApiABC import InboxApiABC
from peek_core_user.server.UserApiABC import UserApiABC

logger = logging.getLogger(__name__)


class LogicEntryHook(PluginLogicEntryHookABC, PluginServerStorageEntryHookABC):
    def __init__(self, *args, **kwargs):
        """ " Constructor"""
        # Call the base classes constructor
        PluginLogicEntryHookABC.__init__(self, *args, **kwargs)

        #: Loaded Objects, This is a list of all objects created when we start
        self._loadedObjects = []

        #: The API object for this plugin
        self._api = None

    def load(self) -> None:
        """Load

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialiastion steps here.

        """

        loadStorageTuples()
        loadPrivateTuples()
        loadPublicTuples()

        logger.debug("Loaded")

    def start(self):
        """Start

        This will be called to start the plugin.
        Start, means what ever we choose to do here. This includes:

        -   Create Controllers

        -   Create payload, observable and tuple action handlers.

        """

        # Create the API
        self._api = ChatApi(self.dbSessionCreator)
        self._loadedObjects.append(self._api)  # For auto shutdown

        userPluginApi = self.platform.getOtherPluginApi("peek_core_user")
        assert isinstance(userPluginApi, UserApiABC), "Expected UserApiABC"

        activeTaskPluginApi = self.platform.getOtherPluginApi(
            "peek_plugin_inbox"
        )
        assert isinstance(
            activeTaskPluginApi, InboxApiABC
        ), "Expected ActiveTaskApiABC"

        self._loadedObjects.extend(
            makeAdminBackendHandlers(self.dbSessionCreator)
        )

        # Tuple Observer, required by main controller
        tupleObservable = makeTupleDataObservableHandler(self.dbSessionCreator)
        self._loadedObjects.append(tupleObservable)

        # Controllers
        taskController = TaskController(activeTaskPluginApi)
        self._loadedObjects.append(taskController)

        mainController = MainController(
            ourApi=self._api,
            dbSessionCreator=self.dbSessionCreator,
            tupleObservable=tupleObservable,
            userPluginApi=userPluginApi,
            taskController=taskController,
        ).start()
        self._loadedObjects.append(mainController)

        self._api.setMainController(mainController)

        # Tuple Action Processor, requires main controller
        self._loadedObjects.append(
            makeTupleActionProcessorHandler(mainController)
        )

        logger.debug("Started")

    def stop(self):
        """Stop

        This method is called by the platform to tell the peek app to shutdown and stop
        everything it's doing
        """
        # Shutdown and dereference all objects we constructed when we started
        while self._loadedObjects:
            self._loadedObjects.pop().shutdown()

        self._api = None

        logger.debug("Stopped")

    def unload(self):
        """Unload

        This method is called after stop is called, to unload any last resources
        before the PLUGIN is unlinked from the platform

        """
        logger.debug("Unloaded")

    @property
    def dbMetadata(self):
        return DeclarativeBase.metadata

    @property
    def publishedServerApi(self) -> Optional[object]:
        """Published Server API

        :return  class that implements the API that can be used by other PLUGINs on this
        platform.
        """
        return self._api



from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import orm
from sqlalchemy.sql.sqltypes import DateTime
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_plugin_base.storage.TypeDecorators import PeekLargeBinary
from peek_plugin_chat._private.PluginNames import chatTuplePrefix
from peek_plugin_chat._private.storage.ChatTuple import ChatTuple
from peek_plugin_chat._private.storage.DeclarativeBase import DeclarativeBase
from peek_plugin_chat.server.ChatApiABC import NewMessage


@addTupleType
class MessageTuple(DeclarativeBase, Tuple):
    __tupleType__ = chatTuplePrefix + "MessageTuple"
    __tablename__ = "MessageTuple"

    id = Column(Integer, primary_key=True, autoincrement=True)

    #: Foreign key to a chat
    chatId = Column(
        Integer, ForeignKey(ChatTuple.id, ondelete="CASCADE"), nullable=False
    )

    # Message details
    message = Column(String(2000), nullable=False)

    priority = Column(Integer, nullable=False)
    PRIORITY_EMERGENCY = NewMessage.PRIORITY_EMERGENCY
    PRIORITY_NORMAL_STICKY = NewMessage.PRIORITY_NORMAL_STICKY
    PRIORITY_NORMAL_FLEETING = NewMessage.PRIORITY_NORMAL_FLEETING

    # User to / from
    fromUserId = Column(String(40), nullable=False)

    # Message state details
    dateTime = Column(DateTime(True), nullable=False)

    onReadPayload = Column(PeekLargeBinary)

    __table_args__ = (
        Index("idx_ChatMsgTuple_chatId", chatId, unique=False),
        Index("idx_ChatMsgTuple_dateTime", dateTime, unique=False),
    )

    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)



from txhttputil.util.ModuleUtil import filterModules


def loadStorageTuples():
    """Load Storage Tables

    This method should be called from the "load()" method of the agent, server, worker
    and client entry hook classes.

    This will register the ORM classes as tuples, allowing them to be serialised and
    deserialized by the vortex.

    """

    for mod in filterModules(__package__, __file__):
        __import__(mod, locals(), globals())



chatPluginName = "peek_plugin_chat"

chatFilt = {"plugin": "peek_plugin_chat"}
chatTuplePrefix = "peek_plugin_chat."
chatObservableName = "peek_plugin_chat"
chatActionProcessorName = "peek_plugin_chat"



from peek_plugin_chat._private.PluginNames import chatFilt
from peek_plugin_chat._private.PluginNames import chatObservableName
from peek_plugin_chat._private.server.tuple_providers.ChatTupleProvider import (
    ChatTupleProvider,
)
from peek_plugin_chat._private.storage.ChatTuple import ChatTuple
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler


def makeTupleDataObservableHandler(ormSessionCreator):
    """ " Make Tuple Data Observable Handler

    This method creates the observable object, registers the tuple providers and then
    returns it.

    :param ormSessionCreator: A function that returns a SQLAlchemy session when called

    :return: An instance of :code:`TupleDataObservableHandler`

    """
    tupleObservable = TupleDataObservableHandler(
        observableName=chatObservableName, additionalFilt=chatFilt
    )

    # Register TupleProviders here
    tupleObservable.addTupleProvider(
        ChatTuple.tupleName(), ChatTupleProvider(ormSessionCreator)
    )

    return tupleObservable



"""timezone aware

Peek Plugin Database Migration Script

Revision ID: a4d6b26d39b7
Revises: 9b0976dcfd53
Create Date: 2017-12-31 14:29:07.288567

"""

# revision identifiers, used by Alembic.
revision = "a4d6b26d39b7"
down_revision = "9b0976dcfd53"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # pl_chat."DeviceInfo"
    op.execute(
        """ALTER TABLE pl_chat."ChatTuple"
                  ALTER COLUMN "lastActivity" TYPE TIMESTAMP WITH TIME ZONE
                  USING "lastActivity" AT TIME ZONE 'UTC'
               """
    )

    # pl_chat."ChatUserTuple"
    op.execute(
        """ALTER TABLE pl_chat."ChatUserTuple"
                  ALTER COLUMN "lastReadDate" TYPE TIMESTAMP WITH TIME ZONE
                  USING "lastReadDate" AT TIME ZONE 'UTC'
               """
    )

    # pl_chat."MessageTuple"
    op.execute(
        """ALTER TABLE pl_chat."MessageTuple"
                  ALTER COLUMN "dateTime" TYPE TIMESTAMP WITH TIME ZONE
                  USING "dateTime" AT TIME ZONE 'UTC'
               """
    )


def downgrade():
    pass



from peek_plugin_base.PeekVortexUtil import peekServerName
from peek_plugin_chat._private.PluginNames import chatFilt
from peek_plugin_chat._private.PluginNames import chatActionProcessorName
from vortex.handler.TupleActionProcessorProxy import TupleActionProcessorProxy


def makeTupleActionProcessorProxy():
    return TupleActionProcessorProxy(
        tupleActionProcessorName=chatActionProcessorName,
        proxyToVortexName=peekServerName,
        additionalFilt=chatFilt,
    )



from typing import Optional, List

from abc import ABCMeta, abstractmethod
from reactivex.subject import Subject


class NewMessageUser:
    """New Message User

    This class represents a user that the message will be sent to.

    """

    def __init__(
        self,
        toUserId: str,
        onReadPayload: Optional[bytes] = None,
        onDeliveredPayload: Optional[bytes] = None,
    ):
        """
        :param toUserId: The peek userId that matches a user in peek_core_user plugin.

        :param onReadPayload: (Optional) The payload that will be delivered locally
            on Peek Server when the user has read the message.

        """
        # To User
        self.toUserId = toUserId
        if not toUserId:
            raise Exception("toUserId is not optional")

        # On Read Payload
        self.onReadPayload = onReadPayload

        #: On Delivered Payload
        self.onDeliveredPayload = onDeliveredPayload


class NewMessage:
    """New Message

    This class represents a new message that another plugin can send to a user.

    """

    # Message priorities

    #:  Emergency priority for message
    PRIORITY_EMERGENCY = 1

    #:  Normal priority for a message, the alert will be fleeting
    PRIORITY_NORMAL_FLEETING = 2

    #:  Normal priority for message, the alert will be sticky
    PRIORITY_NORMAL_STICKY = 3

    def __init__(
        self,
        fromExtUserId: str,
        fromExtUserName: str,
        toUsers: List[NewMessageUser],
        message: str,
        priority: int = PRIORITY_NORMAL_FLEETING,
    ):
        """
        :param fromExtUserId: The external user id of the user sending the message.
            This doesn't have to match a userId in the peek_core_user plugin.

        :param fromExtUserName: The name of the external user (or system) sending the
            message.

        :param toUsers: A list of users to send the message to.

        :param message: The message to send to the user.

        :param priority: The priority of this message, some messages may be emergency
            messages.

        """
        # From User
        self.fromExtUserId = self._required(fromExtUserId, "fromExtUserId")
        self.fromExtUserName = self._required(
            fromExtUserName, "fromExtUserName"
        )

        # To User
        self.toUsers = self._required(toUsers, "toUsers")

        # Message
        self.message = self._required(message, "message")
        self.priority = self._required(priority, "priority")

    def _required(self, val, desc):
        if not val:
            raise Exception("%s is not optional" % desc)

        return val


class ReceivedMessage:
    """Received Message

    This class represents a message sent from a peek user to an external system.

    """

    # Message priorities
    PRIORITY_EMERGENCY = NewMessage.PRIORITY_EMERGENCY
    PRIORITY_NORMAL_FLEETING = NewMessage.PRIORITY_NORMAL_FLEETING
    PRIORITY_NORMAL_STICKY = NewMessage.PRIORITY_NORMAL_STICKY

    def __init__(
        self,
        fromUserId: str,
        allUserIds: List[str],
        message: str,
        priority: int,
    ):
        """
        :param fromUserId: The peek userId sending the message.

        :param allUserIds: All the userIds in this chat.

        :param message: The message sent by the peek user.

        :param priority: The priority of this message sent.

        """
        # From User
        self.fromUserId = fromUserId

        # All Users
        self.allUserIds = allUserIds

        # Message
        self.message = message
        self.priority = priority

    def _required(self, val, desc):
        if not val:
            raise Exception("%s is not optional" % desc)

        return val


class ChatApiABC(metaclass=ABCMeta):
    @abstractmethod
    def sendMessage(self, newMessage: NewMessage) -> None:
        """Send a Message

        Send a new chat message to a user.

        :param newMessage: The definition of the message to send.

        """

    @abstractmethod
    def createChat(self, fromExtUserId: str, toUserIds: List[str]) -> None:
        """Create a Chat

        Send a new chat message to a user.

        :param fromExtUserId: The external userId sending the message
        :param toUserIds: The Peek userIds to send the message to

        """

    @abstractmethod
    def receiveMessages(self, toExtUserId: str) -> Subject:
        """Receive Messages

        Get the observable that will be fired when new messages are received.

        It will be fired with C{ReceivedMessage}

        :param toExtUserId: The external systems userId, that the plugin wants to
            observe messages for. This is just identifier unique to the external
            system.

        :return: A RxJS Observable that will notify observers when a message arrives
            for that external system.

        """



from collections import namedtuple

import pytz
from datetime import datetime, timedelta

import logging
from copy import copy
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from typing import List
from vortex.Payload import Payload
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.VortexFactory import VortexFactory

from peek_plugin_chat._private.PluginNames import chatFilt, chatPluginName
from peek_plugin_chat._private.storage.ChatTuple import ChatTuple
from peek_plugin_chat._private.storage.MessageTuple import MessageTuple
from peek_plugin_inbox.server.InboxApiABC import InboxApiABC, NewTask

logger = logging.getLogger(__name__)

_deliverdPayloadFilt = {"key": "active.task.message.delivered"}
_deliverdPayloadFilt.update(chatFilt)

AddTaskUserTuple = namedtuple(
    "AddTaskUserTuple", ["userId", "onDeliveredPayload"]
)


class TaskController:
    def __init__(self, activeTaskPluginApi: InboxApiABC):
        self._inboxPluginApi = activeTaskPluginApi

        assert isinstance(self._inboxPluginApi, InboxApiABC), (
            "Expected instance of ActiveTaskServerApiABC, received %s"
            % self._inboxPluginApi
        )

        self._deliveredEndpoint = PayloadEndpoint(
            _deliverdPayloadFilt, self._processTaskDelivered
        )

    def shutdown(self):
        self._deliveredEndpoint.shutdown()
        self._deliveredEndpoint = None
        self._soapController = None

    @inlineCallbacks
    def _processTaskDelivered(self, payloadEnvelope: PayloadEnvelope, **kwargs):
        payload = yield payloadEnvelope.decodePayloadDefer()
        logger.debug("_processTaskDelivered called")
        try:
            onDeliverPayload: bytes = payload.tuples
            yield VortexFactory.sendVortexMsgLocally(onDeliverPayload)

        except Exception as e:
            logger.exception(e)

    def _makeUniqueId(self, chatId: int, userId: str):
        return "peek_plugin_chat.new_message.%s.%s" % (userId, chatId)

    def _makeTaskTitle(self, message: MessageTuple):
        if message.priority == MessageTuple.PRIORITY_EMERGENCY:
            return "EMERGENCY SOS CHAT MESSAGE FROM %s" % message.fromUserId

        return "You have a new chat message from %s" % message.fromUserId

    def _makeMessagesRoutePath(self, chatTuple: ChatTuple):
        return "/peek_plugin_chat/messages/%s" % chatTuple.id

    def _notifyBy(self, message: MessageTuple):
        if message.priority == MessageTuple.PRIORITY_EMERGENCY:
            return (
                NewTask.NOTIFY_BY_SMS
                | NewTask.NOTIFY_BY_DEVICE_SOUND
                | NewTask.NOTIFY_BY_DEVICE_DIALOG
            )

        if message.priority == MessageTuple.PRIORITY_NORMAL_STICKY:
            return (
                NewTask.NOTIFY_BY_DEVICE_SOUND | NewTask.NOTIFY_BY_DEVICE_DIALOG
            )

        return NewTask.NOTIFY_BY_DEVICE_SOUND | NewTask.NOTIFY_BY_DEVICE_POPUP

    def _displayPriority(self, message: MessageTuple):
        if message.priority == MessageTuple.PRIORITY_EMERGENCY:
            return NewTask.PRIORITY_DANGER

        return NewTask.PRIORITY_SUCCESS

    def addTask(
        self,
        chat: ChatTuple,
        message: MessageTuple,
        toUsers: List[AddTaskUserTuple],
    ):
        reactor.callLater(0, self._addTask, chat, message, toUsers)

    @inlineCallbacks
    def _addTask(
        self,
        chat: ChatTuple,
        message: MessageTuple,
        toUsers: List[AddTaskUserTuple],
    ):
        try:
            for toUser in toUsers:
                onDeliveredPayloadEnvelope = None
                if toUser.onDeliveredPayload:
                    filt = copy(_deliverdPayloadFilt)
                    filt["chatId"] = chat.id

                    payload = Payload(
                        filt=filt, tuples=toUser.onDeliveredPayload
                    )
                    payloadEnvelope = yield payload.makePayloadEnvelopeDefer()
                    onDeliveredPayloadEnvelope = (
                        yield payloadEnvelope.toVortexMsgDefer()
                    )

                newTask = NewTask(
                    pluginName=chatPluginName,
                    uniqueId=self._makeUniqueId(chat.id, toUser.userId),
                    userId=toUser.userId,
                    title=self._makeTaskTitle(message),
                    description=message.message,
                    displayAs=NewTask.DISPLAY_AS_MESSAGE,
                    displayPriority=self._displayPriority(message),
                    routePath=self._makeMessagesRoutePath(chat),
                    onDeliveredPayloadEnvelope=onDeliveredPayloadEnvelope,
                    autoDelete=NewTask.AUTO_DELETE_ON_SELECT,
                    overwriteExisting=True,
                    notificationRequiredFlags=self._notifyBy(message),
                    autoDeleteDateTime=datetime.now(pytz.utc)
                    + timedelta(minutes=4 * 60),
                )

                yield self._inboxPluginApi.addTask(newTask)

        except Exception as e:
            logger.exception(e)

    def removeTask(self, chatId: int, userId: str):
        reactor.callLater(0, self._removeTask, chatId, userId)

    @inlineCallbacks
    def _removeTask(self, chatId: int, userId: str):
        try:
            yield self._inboxPluginApi.removeTask(
                chatPluginName, self._makeUniqueId(chatId, userId)
            )

        except ValueError:
            # This means it didn't exist.
            pass

        except Exception as e:
            logger.exception(e)



import logging
from typing import Union

from sqlalchemy import desc
from twisted.internet.defer import Deferred

from peek_plugin_chat._private.storage.ChatTuple import ChatTuple
from peek_plugin_chat._private.storage.ChatUserTuple import ChatUserTuple
from peek_plugin_chat._private.storage.MessageTuple import MessageTuple
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

logger = logging.getLogger(__name__)


class ChatTupleProvider(TuplesProviderABC):
    def __init__(self, ormSessionCreator):
        self._ormSessionCreator = ormSessionCreator

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:
        # Potential filters can be placed here.
        chatId = tupleSelector.selector.get("chatId")

        # Potential filters can be placed here.
        userId = tupleSelector.selector.get("userId")

        session = self._ormSessionCreator()
        try:
            chats = []
            # If one chat id has been specified, then just query for one and
            # it's messages.
            if chatId is not None:
                chat = (
                    session.query(ChatTuple)
                    .filter(ChatTuple.id == chatId)
                    .all()
                )

                if chat:
                    chat = chat[0]
                    chat.messages = (
                        session.query(MessageTuple)
                        .filter(MessageTuple.chatId == chatId)
                        .order_by(MessageTuple.dateTime)
                        .all()
                    )

                    chats = [chat]

            # Else the UI is after a list of chats for this user
            else:
                chats = (
                    session.query(ChatTuple)
                    .join(ChatUserTuple, ChatUserTuple.chatId == ChatTuple.id)
                    .filter(ChatUserTuple.userId == userId)
                    .order_by(desc(ChatTuple.lastActivity))
                    .all()
                )

            # Create the vortex message
            msg = (
                Payload(filt, tuples=chats).makePayloadEnvelope().toVortexMsg()
            )

        finally:
            session.close()

        return msg



import logging
from datetime import datetime, timedelta
from typing import List

import pytz
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from twisted.internet.task import LoopingCall

from peek_plugin_base.LoopingCallUtil import peekCatchErrbackWithLogger
from peek_plugin_chat._private.storage.ChatTuple import ChatTuple
from peek_plugin_chat._private.storage.ChatUserTuple import ChatUserTuple
from peek_plugin_chat._private.storage.MessageReadPayloadTuple import (
    MessageReadPayloadTuple,
)
from peek_plugin_chat._private.storage.MessageTuple import MessageTuple
from peek_plugin_chat._private.tuples.ChatUserReadActionTuple import (
    ChatUserReadActionTuple,
)
from peek_plugin_chat._private.tuples.CreateChatActionTuple import (
    CreateChatActionTuple,
)
from peek_plugin_chat._private.tuples.SendMessageActionTuple import (
    SendMessageActionTuple,
)
from peek_plugin_chat.server.ChatApiABC import NewMessage
from peek_core_user.server.UserApiABC import UserApiABC
from vortex.DeferUtil import vortexLogFailure, deferToThreadWrapWithLogger
from vortex.TupleAction import TupleActionABC
from vortex.TupleSelector import TupleSelector
from vortex.VortexFactory import VortexFactory
from vortex.handler.TupleActionProcessor import TupleActionProcessorDelegateABC
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler
from .TaskController import TaskController, AddTaskUserTuple

logger = logging.getLogger(__name__)


class MainController(TupleActionProcessorDelegateABC):
    PROCESS_PERIOD = 600.0  # 10 minutes

    def __init__(
        self,
        dbSessionCreator,
        ourApi,  #: ChatApi,
        userPluginApi: UserApiABC,
        taskController: TaskController,
        tupleObservable: TupleDataObservableHandler,
    ):
        self._ourApi = ourApi
        self._ormSessionCreator = dbSessionCreator
        self._userPluginApi = userPluginApi
        self._taskController = taskController
        self._tupleObserver = tupleObservable

        self._processLoopingCall = LoopingCall(self._deleteOnDateTime)

    def start(self):
        d = self._processLoopingCall.start(self.PROCESS_PERIOD, now=False)
        d.addErrback(vortexLogFailure, logger)
        return self

    def shutdown(self):
        self._processLoopingCall.stop()

    def _notifyOfChatListUpdate(self, userId: str) -> None:
        """Notify of Chat List Update

        Notify the observer of the data it needs to emit to the users.

        :param userId: The user to notify of all chat updates

        """

        self._tupleObserver.notifyOfTupleUpdate(
            TupleSelector(ChatTuple.tupleType(), {"userId": userId})
        )

    def _notifyOfChatUpdate(self, chatId: int) -> None:
        """Notify of Chat Update

        Notify the observer of the data it needs to emit to the users.

        :param chatId: The id of the chat that has been updated.

        """

        self._tupleObserver.notifyOfTupleUpdate(
            TupleSelector(ChatTuple.tupleType(), {"chatId": chatId})
        )

    def processTupleAction(self, tupleAction: TupleActionABC):
        if isinstance(tupleAction, CreateChatActionTuple):
            return self._processCreateChatAction(tupleAction)

        if isinstance(tupleAction, SendMessageActionTuple):
            return self._processSendMessageAction(tupleAction)

        if isinstance(tupleAction, ChatUserReadActionTuple):
            return self._processChatReadAction(tupleAction)

        raise Exception("Unhandled tuple action %s" % tupleAction.tupleType())

    @deferToThreadWrapWithLogger(logger)
    def externalMessageQueued(self, toUserId: str):
        """External Message Sent

        This method is used to notify the main controller when a new message has
        been queued from an external system.
        """

    @deferToThreadWrapWithLogger(logger)
    def _processCreateChatAction(self, action: CreateChatActionTuple):
        """Process Create Chat action by user

        Process updates to the task from the UI.

        """
        session = self._ormSessionCreator()
        allUserIds = action.userIds + [action.fromUserId]

        try:
            self._getOrCreateChatBlocking(allUserIds, [], session)

        finally:
            session.close()

        for userId in allUserIds:
            self._notifyOfChatListUpdate(userId)

    def _getOrCreateChatBlocking(
        self, allUserIds: List[str], extUserIds: List[str], session
    ) -> ChatTuple:
        extUserIds = set(extUserIds)
        usersKey = ",".join(sorted(allUserIds))

        # Check if there is an existing chat
        chatTuple = (
            session.query(ChatTuple)
            .filter(ChatTuple.usersKey == usersKey)
            .all()
        )
        # Convert from the array
        chatTuple = chatTuple[0] if chatTuple else None
        if chatTuple:  # There is an existing one.
            # Bump the chat to the top.
            chatTuple.lastActivity = datetime.now(pytz.utc)

        else:
            # Create the new chat tuple
            chatTuple = ChatTuple()
            chatTuple.lastActivity = datetime.now(pytz.utc)
            chatTuple.usersKey = usersKey
            session.add(chatTuple)

            for userId in allUserIds:
                chatUserTuple = ChatUserTuple()
                chatUserTuple.userId = userId
                chatUserTuple.lastReadDate = datetime.now(pytz.utc)
                chatUserTuple.isUserExternal = userId in extUserIds
                chatUserTuple.userName = userId

                chatTuple.users.append(chatUserTuple)
                session.add(chatUserTuple)
        session.commit()

        return chatTuple

    # -------------------------------------------------------------------------
    # Process send Message Actions
    @inlineCallbacks
    def _processSendMessageAction(self, action: SendMessageActionTuple):
        """Process Task Update

        Process updates to the task from the UI.

        """
        chatTuple, messageTuple = yield self._processSendMessageActionInThread(
            action
        )

        # Tell the API that we've received a message, let it notify who it needs
        self._ourApi.notifyOfReceivedMessage(chatTuple, messageTuple)

        # Get the IDs needed for the updates
        users = [
            chatUser
            for chatUser in chatTuple.users
            if not chatUser.isUserExternal
        ]
        chatId = chatTuple.id

        # Send alerts to the other users.
        alertUsers = [
            AddTaskUserTuple(userId=u.userId, onDeliveredPayload=None)
            for u in users
            if u.userId != action.fromUserId
        ]
        yield self._taskController.addTask(chatTuple, messageTuple, alertUsers)

        for user in users:
            self._notifyOfChatListUpdate(user.userId)

        self._notifyOfChatUpdate(chatId)

    @deferToThreadWrapWithLogger(logger)
    def _processSendMessageActionInThread(self, action: SendMessageActionTuple):
        """Process Task Update

        Process updates to the task from the UI.

        """
        session = self._ormSessionCreator()

        try:
            chatTuple = (
                session.query(ChatTuple)
                .filter(ChatTuple.id == action.chatId)
                .one()
            )

            # Get the chat user for this user, sending a message implies they
            # have read up to date.
            chatUserTuple = list(
                filter(
                    lambda cu: cu.userId == action.fromUserId, chatTuple.users
                )
            )[0]

            # Create the new chat tuple
            messageTuple = MessageTuple()
            messageTuple.chatId = chatTuple.id
            messageTuple.fromUserId = action.fromUserId
            messageTuple.message = action.message
            messageTuple.priority = action.priority
            messageTuple.dateTime = datetime.now(pytz.utc)
            session.add(messageTuple)

            # Update the last activity and lastReadDate for the sender
            chatTuple.lastActivity = datetime.now(pytz.utc)
            chatUserTuple.lastReadDate = datetime.now(pytz.utc)

            # Commit the changes.
            session.commit()
            session.refresh(chatTuple)
            session.refresh(messageTuple)
            session.expunge_all()

            return chatTuple, messageTuple

        finally:
            session.close()

    # -------------------------------------------------------------------------
    # Process Chat Read Actions

    @inlineCallbacks
    def _processChatReadAction(self, action: ChatUserReadActionTuple):
        """Process Create Chat Read action

        Updates the last read date for the user, sends any waiting payloads fo those
        messages

        """
        chatId, userId = yield self._processChatReadActionInThread(action)

        # Remove the unread message if there are any
        yield self._taskController.removeTask(chatId, userId)

    @deferToThreadWrapWithLogger(logger)
    def _processChatReadActionInThread(self, action: ChatUserReadActionTuple):
        """Process Create Chat Read action

        Updates the last read date for the user, sends any waiting payloads fo those
        messages

        """
        session = self._ormSessionCreator()

        try:
            # Find the chat user and update the last read date.
            chatUser = (
                session.query(ChatUserTuple)
                .filter(ChatUserTuple.id == action.chatUserId)
                .one()
            )

            chatId = chatUser.chatId
            userId = chatUser.userId

            chatUser.lastReadDate = action.readDateTime

            # Send any onRead payloads that are required, and cleanup.
            msgPayloads = (
                session.query(MessageReadPayloadTuple)
                .join(
                    MessageTuple,
                    MessageTuple.id == MessageReadPayloadTuple.messageId,
                )
                .filter(MessageTuple.dateTime <= action.readDateTime)
                .filter(MessageReadPayloadTuple.chatUserId == chatUser.id)
                .all()
            )

            for msgPayload in msgPayloads:
                if msgPayload.onReadPayload:
                    reactor.callLater(
                        0,
                        VortexFactory.sendVortexMsgLocally,
                        msgPayload.onReadPayload,
                    )

                session.delete(msgPayload)

            session.commit()

            return chatId, userId

        finally:
            session.close()

    # -------------------------------------------------------------------------
    # Send Message From External User

    @inlineCallbacks
    def sendMessageFromExternalUser(self, newMessage: NewMessage):
        """Send Message From External User

        This method handles the messages sent via the API

        :param newMessage: A new message to send, in the APIs simple format.
        :return: None

        """
        (
            chatTuple,
            messageTuple,
            allUserIds,
        ) = yield self._sendMessageFromExternalUserInThread(newMessage)

        # Send alerts to the other users.
        toUsers = [
            AddTaskUserTuple(
                userId=u.toUserId, onDeliveredPayload=u.onDeliveredPayload
            )
            for u in newMessage.toUsers
        ]
        self._taskController.addTask(chatTuple, messageTuple, toUsers)

        for userId in allUserIds:
            self._notifyOfChatListUpdate(userId)

        self._notifyOfChatUpdate(chatTuple.id)

    @deferToThreadWrapWithLogger(logger)
    def _sendMessageFromExternalUserInThread(self, newMessage: NewMessage):
        """Send Message From External User

        This method handles the messages sent via the API

        :param newMessage: A new message to send, in the APIs simple format.
        :return: None

        """

        session = self._ormSessionCreator()
        try:
            # Create an array of all users in the chat
            allUserIds = [nmu.toUserId for nmu in newMessage.toUsers]
            allUserIds += [newMessage.fromExtUserId]

            # Get or create the chat tuple
            chatTuple = self._getOrCreateChatBlocking(
                allUserIds, [newMessage.fromExtUserId], session
            )

            # Ensure that the external user is set as an external user
            extChatUser = list(
                filter(
                    lambda cu: cu.userId == newMessage.fromExtUserId,
                    chatTuple.users,
                )
            )[0]
            extChatUser.isUserExternal = True

            # Create the new chat tuple
            messageTuple = MessageTuple()
            messageTuple.chatId = chatTuple.id
            messageTuple.fromUserId = newMessage.fromExtUserId
            messageTuple.message = newMessage.message
            messageTuple.priority = newMessage.priority
            messageTuple.dateTime = datetime.now(pytz.utc)
            session.add(messageTuple)

            # Create a map of userIds to IDs for the user read payloads
            chatUserByUserId = {cu.userId: cu for cu in chatTuple.users}

            # Create the read payload tuples
            for toUser in newMessage.toUsers:
                if not toUser.onReadPayload:
                    continue

                readPayload = MessageReadPayloadTuple()
                readPayload.message = messageTuple
                readPayload.chatUser = chatUserByUserId[toUser.toUserId]
                readPayload.onReadPayload = toUser.onReadPayload
                session.add(readPayload)

            session.commit()
            session.refresh(chatTuple)
            session.refresh(messageTuple)
            session.expunge_all()

            return chatTuple, messageTuple, allUserIds

        finally:
            session.close()

    # -------------------------------------------------------------------------
    # Create Chat

    @deferToThreadWrapWithLogger(logger)
    def createChat(self, fromExtUserId: str, toUserIds: List[str]) -> None:
        """Create Chat

        Used by the API

        """

        session = self._ormSessionCreator()
        try:
            # Create an array of all users in the chat
            allUserIds = list(toUserIds)
            allUserIds += [fromExtUserId]

            # Get or create the chat tuple
            chatTuple = self._getOrCreateChatBlocking(
                allUserIds, [fromExtUserId], session
            )
            chatId = chatTuple.id

            session.commit()

        finally:
            session.close()

        for userId in allUserIds:
            self._notifyOfChatListUpdate(userId)

        self._notifyOfChatUpdate(chatId)

    # -------------------------------------------------------
    # Delete Old Messages
    # -------------------------------------------------------

    @peekCatchErrbackWithLogger(logger)
    @deferToThreadWrapWithLogger(logger)
    def _deleteOnDateTime(self):
        session = self._ormSessionCreator()
        # chatUsersIdsToUpdate = set()

        try:
            expiredMessageDate = datetime.now(pytz.utc) - timedelta(days=2)
            expiredChatDate = datetime.now(pytz.utc) - timedelta(days=5)

            # Query for the chats to delete
            chatUsersIdsToUpdate = set(
                map(
                    lambda i: i[0],
                    session.query(ChatUserTuple.userId)
                    .join(ChatTuple, ChatTuple.id == ChatUserTuple.chatId)
                    .filter(ChatTuple.lastActivity < expiredChatDate),
                )
            )

            # No updates required for message expiration, just delete them.
            (
                session.query(MessageTuple)
                .filter(MessageTuple.dateTime < expiredMessageDate)
                .delete()
            )

            # Delete Chats
            (
                session.query(ChatTuple)
                .filter(ChatTuple.lastActivity < expiredChatDate)
                .delete()
            )

            session.commit()

        finally:
            session.close()

        for userId in chatUsersIdsToUpdate:
            self._notifyOfChatListUpdate(userId)



"""Added read payload constraint

Peek Plugin Database Migration Script

Revision ID: 52f3163b45de
Revises: a4d6b26d39b7
Create Date: 2018-01-27 11:25:04.436708

"""

# revision identifiers, used by Alembic.
revision = "52f3163b45de"
down_revision = "a4d6b26d39b7"
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        "MessageReadPayloadTuple_chatUserId_fkey",
        "MessageReadPayloadTuple",
        schema="pl_chat",
        type_="foreignkey",
    )
    op.create_foreign_key(
        None,
        "MessageReadPayloadTuple",
        "ChatUserTuple",
        ["chatUserId"],
        ["id"],
        source_schema="pl_chat",
        referent_schema="pl_chat",
        ondelete="CASCADE",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        None, "MessageReadPayloadTuple", schema="pl_chat", type_="foreignkey"
    )
    op.create_foreign_key(
        "MessageReadPayloadTuple_chatUserId_fkey",
        "MessageReadPayloadTuple",
        "ChatUserTuple",
        ["chatUserId"],
        ["id"],
        source_schema="pl_chat",
        referent_schema="pl_chat",
    )
    # ### end Alembic commands ###



"""Truncate chats

Peek Plugin Database Migration Script

Revision ID: 9b0976dcfd53
Revises: 5288d02e94e4
Create Date: 2017-12-27 10:03:38.828892

"""

# revision identifiers, used by Alembic.
revision = "9b0976dcfd53"
down_revision = "5288d02e94e4"
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.execute(' TRUNCATE TABLE pl_chat."ChatTuple" CASCADE ')


def downgrade():
    pass



from typing import List

from peek_plugin_chat._private.PluginNames import chatTuplePrefix
from vortex.Tuple import addTupleType, TupleField
from vortex.TupleAction import TupleActionABC


@addTupleType
class CreateChatActionTuple(TupleActionABC):
    __tupleType__ = chatTuplePrefix + "CreateChatActionTuple"

    userIds = TupleField(typingType=List)
    fromUserId = TupleField(typingType=str)
